/*
 * Copyright 2011 maru project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.maru.dog.bind;

import static org.maru.common.type.PrimitiveWrapperTypes.isPrimitiveOrWrapperType;
import static org.maru.common.util.ConditionUtil.isNotNull;
import static org.maru.dog.bind.TypeSafeChecker.verifyTypeSafe;

import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.maru.common.Key;
import org.maru.common.KeyGen;
import org.maru.dog.core.Builder;
import org.maru.dog.core.Configuration;
import org.maru.dog.core.Definition;
import org.maru.dog.core.MarkedPoint;

public abstract class AbstractBinderBuilder<T, K> implements Builder {

    protected final T target;
    protected final K[] inputs;
    protected final InstanceBinder<T, K> instanceBinder;
    protected final List<Configuration> configurations;

    public AbstractBinderBuilder(List<Configuration> configurations, InstanceBinder<T, K> instanceBinder,
            T target, K... inputs) {
        this.configurations = configurations;
        this.target = target;
        this.inputs = inputs;
        this.instanceBinder = instanceBinder;
    }

    public InstanceBinder<T, K> getInstanceBinder() {
        return this.instanceBinder;
    }

    @Override
    public Builder build() {
        Map<Key<?>,Definition<K>> inputDefinitions = buildInputMarkedPoints();
        Definition<T> targetDefinition = buildTargetMarkedPoints();
        validateAndCreateInstanceBinder(targetDefinition, inputDefinitions);
        return this;
    }

    /**
     * Builds {@link InputMarkedPoint}
     * @return input marked points.
     */
    @SuppressWarnings("unchecked")
    protected Map<Key<?>, Definition<K>>buildInputMarkedPoints() {
        Map<Key<?>,Definition<K>> definitions = new HashMap<Key<?>,Definition<K>>();
        for (K input : inputs) {
            Class<K> inputClass = (Class<K>) input.getClass();
            if (isPrimitiveOrWrapperType(inputClass)) {
                continue;
            }
            Key<?> key = KeyGen.getKey(inputClass);

            Definition<K> definition = createAndStoreInputDefinition(key, inputClass);
            if (!definitions.containsKey(key)) {
                definitions.put(key, definition);
            }
        }
        return definitions;
    }

    /**
     * Builds {@link TargetMarkedPoint}
     * @return target marked points.
     */
    @SuppressWarnings("unchecked")
    protected Definition<T> buildTargetMarkedPoints() {
        Key<?> key = KeyGen.getKey(target.getClass());
        Class<T> targetType = (Class<T>) target.getClass();

        return createAndStoreTargetDefinition(key, targetType);
    }

    /**
     * Creates {@code InstanceBinderImpl}
     *
     * @param targetDefinition
     * @param inputDefinitions
     */
    protected void validateAndCreateInstanceBinder(Definition<T> targetDefinition, Map<Key<?>,Definition<K>> inputDefinitions) {

        Key<?> targetTypeKey = KeyGen.getKey(targetDefinition.getType());
        for (Entry<Key<?>, MarkedPoint<T>> entry : ((BindingDefinition<T>)targetDefinition).getMarkedPoints().entrySet()) {
            Key<?> targetBoundKey = entry.getKey();
            TargetMarkedPoint<T> targetMarkedPoint = (TargetMarkedPoint<T>)entry.getValue();
            Class<?> inputClass = null;

            //  get binding configuration if it's defined.
            BindingConfiguration bindingConfiguration = searchBindingConfiguration(targetTypeKey, targetBoundKey);
            if (isNotNull(bindingConfiguration)) {
                inputClass = bindingConfiguration.getInputClass();
                Key<?> bindKey = isNotNull(bindingConfiguration.getBindName()) ? KeyGen.getKey(bindingConfiguration.getBindName()) : targetBoundKey;

                // if binding context is created, then go to next loop
                BindingContext<T, K> context = createBindingContext(targetMarkedPoint, bindKey, inputDefinitions, inputClass, bindingConfiguration);

                if (isNotNull(context)) {
                    MemberBinder<T, K> memberBinder = createMemberBinder(context);
                    instanceBinder.addMemberBinder(memberBinder);
                }
                continue;
            }

            inputClass = targetMarkedPoint.getInputClass();

            BindingContext<T, K> context = createBindingContext(targetMarkedPoint, targetBoundKey, inputDefinitions, inputClass, null);
            if (isNotNull(context)) {
                MemberBinder<T, K> memberBinder = createMemberBinder(context);
                instanceBinder.addMemberBinder(memberBinder);
            }

        }
    }

    /**
     * search and get defined {@link BindingConfiguration} if it's defined in LOCAL_CONFIGURATION_INFOS
     *
     * @param targetTypeKey
     * @param targetBoundKey
     * @return {@link BindingConfiguration}
     */
    private BindingConfiguration searchBindingConfiguration(Key<?> targetTypeKey, Key<?> targetBoundKey) {

        BindingConfiguration configuration = null;
        if (isNotNull(configurations) && configurations.size() > 0) {
            for (Configuration c : configurations) {
                Key<?> configKey = KeyGen.getKey(((BindingConfiguration)c).getBoundName());
                if (targetBoundKey.equals(configKey)) {
                    configuration = (BindingConfiguration)c;
                    break;
                }
            }
        }
        return configuration;
    }

    protected abstract Definition<K> createAndStoreInputDefinition(Key<?> key, Class<K> inputClass);

    protected abstract Definition<T> createAndStoreTargetDefinition(Key<?> key, Class<T> targetType);

    /**
     * Creates and returns {@link BindingContext}
     */
    @SuppressWarnings("unchecked")
    protected static <T, K> BindingContext<T, K> createBindingContext(TargetMarkedPoint<T> targetMarkedPoint, Key<?> targetBoundKey,
            Map<Key<?>,Definition<K>> inputDefinitions, Class<?> inputClass, BindingConfiguration configuration) {

        BindingContext<T, K> bindingContext = null;
        if (isNotNull(inputClass)) {
            Key<?> inputClassKey = KeyGen.getKey(inputClass);
            if (inputDefinitions.containsKey(inputClassKey)) {
                Definition<?> inputDefinition = inputDefinitions.get(inputClassKey);

                if (((BindingDefinition<K>)inputDefinition).containsMarkedPoint(targetBoundKey)) {
                    InputMarkedPoint<K> inputMarkedPoint = (InputMarkedPoint<K>)((BindingDefinition<K>)inputDefinition).getMarkedPoint(targetBoundKey);

                    // type safe check.
                    verifyTypeSafe(inputMarkedPoint, targetMarkedPoint);

                    bindingContext = new BindingContext<T, K>(targetMarkedPoint, inputMarkedPoint, configuration);
                }
            }
        } else {
            int duplicateCount = 0;
            InputMarkedPoint<K> inputMarkedPoint = null;
            for (Entry<Key<?>, Definition<K>> inputEntry : inputDefinitions.entrySet()) {
                for (Entry<Key<?>, MarkedPoint<K>> definedEntry : ((BindingDefinition<K>) inputEntry.getValue()).getMarkedPoints().entrySet()) {
                    InputMarkedPoint<K> tmpInputMarkedPoint = (InputMarkedPoint<K>)definedEntry.getValue();
                    if (tmpInputMarkedPoint.getName().equals(targetMarkedPoint.getName())) {
                        if (duplicateCount == 0) {
                            inputMarkedPoint = tmpInputMarkedPoint;
                        }
                        duplicateCount++;
                    }
                }
            }

            if (duplicateCount > 1) {
                throw new IllegalArgumentException("The name [" + inputMarkedPoint.getName() + "] is duplicate.");
            }

            if (duplicateCount == 1) {
                // type safe check.
                verifyTypeSafe(inputMarkedPoint, targetMarkedPoint);

                bindingContext = new BindingContext<T, K>(targetMarkedPoint, inputMarkedPoint, configuration);
            }
        }
        return bindingContext;
    }

    protected static <T, K> MemberBinder<T, K> createMemberBinder(BindingContext<T, K> context) {
        Member member = context.getTargetMarkedPoint().getMember();

        if (member instanceof Field) {
            return  new FieldBinder<T, K>(context);
        } else if (member instanceof Method) {
            return new AccessorMethodBinder<T, K>(context);
        } else {
            throw new IllegalArgumentException("Fails to solve the field or method binder.");
        }

    }

    public T getTargetInstance() {
        return target;
    }

    public K[] getInputInstances() {
        return inputs;
    }

}
